"use strict";
/**
 * Copyright (c) 2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * Inspired by https://github.com/dgasmith/gau2grid.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.computeIsocontourValues = exports.createSphericalCollocationGrid = void 0;
var tslib_1 = require("tslib");
var util_1 = require("../../mol-data/util");
var geometry_1 = require("../../mol-math/geometry");
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var mol_task_1 = require("../../mol-task");
var array_1 = require("../../mol-util/array");
var collocation_1 = require("./collocation");
var compute_1 = require("./gpu/compute");
function createSphericalCollocationGrid(params, webgl) {
    var _this = this;
    return mol_task_1.Task.create('Spherical Collocation Grid', function (ctx) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
        var centers, cParams, matrix;
        return tslib_1.__generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    centers = params.basis.atoms.map(function (a) { return a.center; });
                    cParams = {
                        grid: initBox(centers, params.gridSpacing, params.boxExpand),
                        basis: params.basis,
                        alphaOrbitals: params.alphaOrbitals,
                        cutoffThreshold: params.cutoffThreshold,
                        sphericalOrder: params.sphericalOrder
                    };
                    if (!compute_1.canComputeAlphaOrbitalsOnGPU(webgl)) return [3 /*break*/, 1];
                    // console.time('gpu');
                    matrix = compute_1.gpuComputeAlphaOrbitalsGridValues(webgl, cParams);
                    return [3 /*break*/, 3];
                case 1: return [4 /*yield*/, collocation_1.sphericalCollocation(cParams, ctx)];
                case 2:
                    // console.time('cpu');
                    matrix = _a.sent();
                    _a.label = 3;
                case 3: return [2 /*return*/, createCubeGrid(cParams.grid, matrix, [0, 1, 2], !params.doNotComputeIsovalues)];
            }
        });
    }); });
}
exports.createSphericalCollocationGrid = createSphericalCollocationGrid;
var BohrToAngstromFactor = 0.529177210859;
function createCubeGrid(gridInfo, values, axisOrder, computeIsovalues) {
    var boxSize = geometry_1.Box3D.size(linear_algebra_1.Vec3(), gridInfo.box);
    var boxOrigin = linear_algebra_1.Vec3.clone(gridInfo.box.min);
    linear_algebra_1.Vec3.scale(boxSize, boxSize, BohrToAngstromFactor);
    linear_algebra_1.Vec3.scale(boxOrigin, boxOrigin, BohrToAngstromFactor);
    var scale = linear_algebra_1.Mat4.fromScaling(linear_algebra_1.Mat4(), linear_algebra_1.Vec3.div(linear_algebra_1.Vec3(), boxSize, linear_algebra_1.Vec3.sub(linear_algebra_1.Vec3(), gridInfo.dimensions, linear_algebra_1.Vec3.create(1, 1, 1))));
    var translate = linear_algebra_1.Mat4.fromTranslation(linear_algebra_1.Mat4(), boxOrigin);
    var matrix = linear_algebra_1.Mat4.mul(linear_algebra_1.Mat4(), translate, scale);
    var grid = {
        transform: { kind: 'matrix', matrix: matrix },
        cells: linear_algebra_1.Tensor.create(linear_algebra_1.Tensor.Space(gridInfo.dimensions, axisOrder, Float32Array), values),
        stats: {
            min: array_1.arrayMin(values),
            max: array_1.arrayMax(values),
            mean: array_1.arrayMax(values),
            sigma: array_1.arrayRms(values),
        },
    };
    // TODO: when using GPU rendering, the cumulative sum can be computed
    //       along the ray on the fly?
    var isovalues;
    if (computeIsovalues) {
        isovalues = computeIsocontourValues(values, 0.85);
    }
    return { grid: grid, isovalues: isovalues };
}
function initBox(geometry, spacing, expand) {
    var count = geometry.length;
    var box = geometry_1.Box3D.expand(geometry_1.Box3D(), geometry_1.Box3D.fromVec3Array(geometry_1.Box3D(), geometry), linear_algebra_1.Vec3.create(expand, expand, expand));
    var size = geometry_1.Box3D.size(linear_algebra_1.Vec3(), box);
    var spacingThresholds = typeof spacing === 'number' ? [[0, spacing]] : tslib_1.__spreadArrays(spacing);
    spacingThresholds.sort(function (a, b) { return b[0] - a[0]; });
    var s = 0.4;
    for (var i = 0; i <= spacingThresholds.length; i++) {
        s = spacingThresholds[i][1];
        if (spacingThresholds[i][0] <= count)
            break;
    }
    var dimensions = linear_algebra_1.Vec3.ceil(linear_algebra_1.Vec3(), linear_algebra_1.Vec3.scale(linear_algebra_1.Vec3(), size, 1 / s));
    return {
        box: box,
        dimensions: dimensions,
        size: size,
        npoints: dimensions[0] * dimensions[1] * dimensions[2],
        delta: linear_algebra_1.Vec3.div(linear_algebra_1.Vec3(), size, linear_algebra_1.Vec3.subScalar(linear_algebra_1.Vec3(), dimensions, 1)),
    };
}
function computeIsocontourValues(input, cumulativeThreshold) {
    var weightSum = 0;
    for (var i = 0, _i = input.length; i < _i; i++) {
        var v = input[i];
        var w = v * v;
        weightSum += w;
    }
    var avgWeight = weightSum / input.length;
    var minWeight = 3 * avgWeight;
    // do not try to identify isovalues for degenerate data
    // e.g. all values are almost same
    if (Math.abs(avgWeight - input[0] * input[0]) < 1e-5) {
        return { negative: void 0, positive: void 0 };
    }
    var size = 0;
    while (true) {
        var csum = 0;
        size = 0;
        for (var i = 0, _i = input.length; i < _i; i++) {
            var v = input[i];
            var w = v * v;
            if (w >= minWeight) {
                csum += w;
                size++;
            }
        }
        if (csum / weightSum > cumulativeThreshold) {
            break;
        }
        minWeight -= avgWeight;
    }
    var values = new Float32Array(size);
    var weights = new Float32Array(size);
    var indices = new Int32Array(size);
    var o = 0;
    for (var i = 0, _i = input.length; i < _i; i++) {
        var v = input[i];
        var w = v * v;
        if (w >= minWeight) {
            values[o] = v;
            weights[o] = w;
            indices[o] = o;
            o++;
        }
    }
    util_1.sortArray(indices, function (indices, i, j) { return weights[indices[j]] - weights[indices[i]]; });
    var cweight = 0, cutoffIndex = 0;
    for (var i = 0; i < size; i++) {
        cweight += weights[indices[i]];
        if (cweight / weightSum >= cumulativeThreshold) {
            cutoffIndex = i;
            break;
        }
    }
    var positive = Number.POSITIVE_INFINITY, negative = Number.NEGATIVE_INFINITY;
    for (var i = 0; i < cutoffIndex; i++) {
        var v = values[indices[i]];
        if (v > 0) {
            if (v < positive)
                positive = v;
        }
        else if (v < 0) {
            if (v > negative)
                negative = v;
        }
    }
    return {
        negative: negative !== Number.NEGATIVE_INFINITY ? negative : void 0,
        positive: positive !== Number.POSITIVE_INFINITY ? positive : void 0,
    };
}
exports.computeIsocontourValues = computeIsocontourValues;
//# sourceMappingURL=cubes.js.map