"use strict";
/**
 * Copyright (c) 2019-2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.RootStructureDefinition = void 0;
var tslib_1 = require("tslib");
var structure_1 = require("../../mol-model/structure");
var string_1 = require("../../mol-util/string");
var geometry_1 = require("../../mol-math/geometry");
var param_definition_1 = require("../../mol-util/param-definition");
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var symmetry_1 = require("../../mol-model/structure/model/properties/symmetry");
var objects_1 = require("../objects");
var symmetry_2 = require("../../mol-model-formats/structure/property/symmetry");
var RootStructureDefinition;
(function (RootStructureDefinition) {
    function getParams(model, defaultValue) {
        var symmetry = model && symmetry_2.ModelSymmetry.Provider.get(model);
        var assemblyIds = symmetry ? symmetry.assemblies.map(function (a) { return [a.id, a.id + ": " + string_1.stringToWords(a.details)]; }) : [];
        var showSymm = !symmetry ? true : !geometry_1.SpacegroupCell.isZero(symmetry.spacegroup.cell);
        var operatorOptions = [];
        if (symmetry) {
            var operators = symmetry.spacegroup.operators;
            for (var i = 0, il = operators.length; i < il; i++) {
                operatorOptions.push([i, i + 1 + ": " + geometry_1.Spacegroup.getOperatorXyz(operators[i])]);
            }
        }
        var asymIdsOptions = [];
        if (model) {
            model.properties.structAsymMap.forEach(function (v) {
                var label = v.id === v.auth_id ? v.id : v.id + " [auth " + v.auth_id + "]";
                asymIdsOptions.push([v.id, label]);
            });
        }
        var modes = {
            auto: param_definition_1.ParamDefinition.EmptyGroup(),
            model: param_definition_1.ParamDefinition.EmptyGroup(),
            assembly: param_definition_1.ParamDefinition.Group({
                id: param_definition_1.ParamDefinition.Optional(model
                    ? param_definition_1.ParamDefinition.Select(assemblyIds.length ? assemblyIds[0][0] : '', assemblyIds, { label: 'Asm Id', description: 'Assembly Id' })
                    : param_definition_1.ParamDefinition.Text('', { label: 'Asm Id', description: 'Assembly Id (use empty for the 1st assembly)' }))
            }, { isFlat: true }),
            'symmetry-mates': param_definition_1.ParamDefinition.Group({
                radius: param_definition_1.ParamDefinition.Numeric(5, { min: 0, max: 50, step: 1 })
            }, { isFlat: true }),
            'symmetry': param_definition_1.ParamDefinition.Group({
                ijkMin: param_definition_1.ParamDefinition.Vec3(linear_algebra_1.Vec3.create(-1, -1, -1), { step: 1 }, { label: 'Min IJK', fieldLabels: { x: 'I', y: 'J', z: 'K' } }),
                ijkMax: param_definition_1.ParamDefinition.Vec3(linear_algebra_1.Vec3.create(1, 1, 1), { step: 1 }, { label: 'Max IJK', fieldLabels: { x: 'I', y: 'J', z: 'K' } })
            }, { isFlat: true }),
            'symmetry-assembly': param_definition_1.ParamDefinition.Group({
                generators: param_definition_1.ParamDefinition.ObjectList({
                    operators: param_definition_1.ParamDefinition.ObjectList({
                        index: param_definition_1.ParamDefinition.Select(0, operatorOptions),
                        shift: param_definition_1.ParamDefinition.Vec3(linear_algebra_1.Vec3(), { step: 1 }, { label: 'IJK', fieldLabels: { x: 'I', y: 'J', z: 'K' } })
                    }, function (e) { return e.index + 1 + "_" + e.shift.map(function (a) { return a + 5; }).join(''); }, {
                        defaultValue: []
                    }),
                    asymIds: param_definition_1.ParamDefinition.MultiSelect([], asymIdsOptions)
                }, function (e) { return e.asymIds.length + " asym ids, " + e.operators.length + " operators"; }, {
                    defaultValue: []
                })
            }, { isFlat: true })
        };
        var options = [];
        if (defaultValue === 'auto') {
            options.push(['auto', 'Auto']);
        }
        options.push(['model', 'Model']);
        if (assemblyIds.length > 0) {
            options.push(['assembly', 'Assembly']);
        }
        if (showSymm) {
            options.push(['symmetry-mates', 'Symmetry Mates']);
            options.push(['symmetry', 'Symmetry (indices)']);
            options.push(['symmetry-assembly', 'Symmetry (assembly)']);
        }
        return {
            type: param_definition_1.ParamDefinition.MappedStatic(defaultValue || 'model', modes, { options: options })
        };
    }
    RootStructureDefinition.getParams = getParams;
    function canAutoUpdate(oldParams, newParams) {
        if (newParams.name === 'symmetry-assembly' || (newParams.name === 'symmetry' && oldParams.name === 'symmetry'))
            return false;
        return true;
    }
    RootStructureDefinition.canAutoUpdate = canAutoUpdate;
    function buildAssembly(plugin, ctx, model, id) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var asm, symmetry, base, label, s, props;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        asm = void 0;
                        symmetry = symmetry_2.ModelSymmetry.Provider.get(model);
                        // if no id is specified, use the 1st assembly.
                        if (!id && symmetry && symmetry.assemblies.length !== 0) {
                            id = symmetry.assemblies[0].id;
                        }
                        if (!symmetry || symmetry.assemblies.length === 0) {
                            plugin.log.warn("Model '" + model.entryId + "' has no assembly, returning model structure.");
                        }
                        else {
                            asm = symmetry_1.Symmetry.findAssembly(model, id || '');
                            if (!asm) {
                                plugin.log.warn("Model '" + model.entryId + "' has no assembly called '" + id + "', returning model structure.");
                            }
                        }
                        base = structure_1.Structure.ofModel(model);
                        if (!asm) {
                            label = { label: 'Model', description: structure_1.Structure.elementDescription(base) };
                            return [2 /*return*/, new objects_1.PluginStateObject.Molecule.Structure(base, label)];
                        }
                        id = asm.id;
                        return [4 /*yield*/, structure_1.StructureSymmetry.buildAssembly(base, id).runInContext(ctx)];
                    case 1:
                        s = _a.sent();
                        props = { label: "Assembly " + id, description: structure_1.Structure.elementDescription(s) };
                        return [2 /*return*/, new objects_1.PluginStateObject.Molecule.Structure(s, props)];
                }
            });
        });
    }
    function buildSymmetry(ctx, model, ijkMin, ijkMax) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var base, s, props;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        base = structure_1.Structure.ofModel(model);
                        return [4 /*yield*/, structure_1.StructureSymmetry.buildSymmetryRange(base, ijkMin, ijkMax).runInContext(ctx)];
                    case 1:
                        s = _a.sent();
                        props = { label: "Symmetry [" + ijkMin + "] to [" + ijkMax + "]", description: structure_1.Structure.elementDescription(s) };
                        return [2 /*return*/, new objects_1.PluginStateObject.Molecule.Structure(s, props)];
                }
            });
        });
    }
    function buildSymmetryMates(ctx, model, radius) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var base, s, props;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        base = structure_1.Structure.ofModel(model);
                        return [4 /*yield*/, structure_1.StructureSymmetry.builderSymmetryMates(base, radius).runInContext(ctx)];
                    case 1:
                        s = _a.sent();
                        props = { label: "Symmetry Mates", description: structure_1.Structure.elementDescription(s) };
                        return [2 /*return*/, new objects_1.PluginStateObject.Molecule.Structure(s, props)];
                }
            });
        });
    }
    function buildSymmetryAssembly(ctx, model, generators, symmetry) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var base, s, props;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        base = structure_1.Structure.ofModel(model);
                        return [4 /*yield*/, structure_1.StructureSymmetry.buildSymmetryAssembly(base, generators, symmetry).runInContext(ctx)];
                    case 1:
                        s = _a.sent();
                        props = { label: "Symmetry Assembly", description: structure_1.Structure.elementDescription(s) };
                        return [2 /*return*/, new objects_1.PluginStateObject.Molecule.Structure(s, props)];
                }
            });
        });
    }
    function create(plugin, ctx, model, params) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var symmetry, s, s;
            return tslib_1.__generator(this, function (_a) {
                symmetry = symmetry_2.ModelSymmetry.Provider.get(model);
                if (!symmetry || !params || params.name === 'model') {
                    s = structure_1.Structure.ofModel(model);
                    return [2 /*return*/, new objects_1.PluginStateObject.Molecule.Structure(s, { label: 'Model', description: structure_1.Structure.elementDescription(s) })];
                }
                if (params.name === 'auto') {
                    if (symmetry.assemblies.length === 0) {
                        s = structure_1.Structure.ofModel(model);
                        return [2 /*return*/, new objects_1.PluginStateObject.Molecule.Structure(s, { label: 'Model', description: structure_1.Structure.elementDescription(s) })];
                    }
                    else {
                        return [2 /*return*/, buildAssembly(plugin, ctx, model)];
                    }
                }
                if (params.name === 'assembly') {
                    return [2 /*return*/, buildAssembly(plugin, ctx, model, params.params.id)];
                }
                if (params.name === 'symmetry') {
                    return [2 /*return*/, buildSymmetry(ctx, model, params.params.ijkMin, params.params.ijkMax)];
                }
                if (params.name === 'symmetry-mates') {
                    return [2 /*return*/, buildSymmetryMates(ctx, model, params.params.radius)];
                }
                if (params.name === 'symmetry-assembly') {
                    return [2 /*return*/, buildSymmetryAssembly(ctx, model, params.params.generators, symmetry)];
                }
                throw new Error("Unknown represetation type: " + params.name);
            });
        });
    }
    RootStructureDefinition.create = create;
})(RootStructureDefinition = exports.RootStructureDefinition || (exports.RootStructureDefinition = {}));
//# sourceMappingURL=root-structure.js.map