"use strict";
/**
 * Copyright (c) 2018-2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.InterUnitBondCylinderMeshVisual = exports.InterUnitBondCylinderImpostorVisual = exports.InterUnitBondCylinderVisual = exports.InterUnitBondCylinderParams = void 0;
var tslib_1 = require("tslib");
var param_definition_1 = require("../../../mol-util/param-definition");
var structure_1 = require("../../../mol-model/structure");
var mesh_1 = require("../../../mol-geo/geometry/mesh/mesh");
var linear_algebra_1 = require("../../../mol-math/linear-algebra");
var mol_util_1 = require("../../../mol-util");
var link_1 = require("./util/link");
var complex_visual_1 = require("../complex-visual");
var types_1 = require("../../../mol-model/structure/model/types");
var bond_1 = require("./util/bond");
var geometry_1 = require("../../../mol-math/geometry");
var cylinders_1 = require("../../../mol-geo/geometry/cylinders/cylinders");
var tmpRefPosBondIt = new structure_1.Bond.ElementBondIterator();
function setRefPosition(pos, structure, unit, index) {
    tmpRefPosBondIt.setElement(structure, unit, index);
    while (tmpRefPosBondIt.hasNext) {
        var bA = tmpRefPosBondIt.move();
        bA.otherUnit.conformation.position(bA.otherUnit.elements[bA.otherIndex], pos);
        return pos;
    }
    return null;
}
var tmpRef = linear_algebra_1.Vec3();
function getInterUnitBondCylinderBuilderProps(structure, theme, props) {
    var locE = structure_1.StructureElement.Location.create(structure);
    var locB = structure_1.Bond.Location(structure, undefined, undefined, structure, undefined, undefined);
    var bonds = structure.interUnitBonds;
    var edgeCount = bonds.edgeCount, edges = bonds.edges;
    var sizeFactor = props.sizeFactor, sizeAspectRatio = props.sizeAspectRatio;
    var delta = linear_algebra_1.Vec3();
    var radius = function (edgeIndex) {
        var b = edges[edgeIndex];
        locB.aUnit = structure.unitMap.get(b.unitA);
        locB.aIndex = b.indexA;
        locB.bUnit = structure.unitMap.get(b.unitB);
        locB.bIndex = b.indexB;
        return theme.size.size(locB) * sizeFactor;
    };
    var radiusA = function (edgeIndex) {
        var b = edges[edgeIndex];
        locE.unit = structure.unitMap.get(b.unitA);
        locE.element = locE.unit.elements[b.indexA];
        return theme.size.size(locE) * sizeFactor;
    };
    var radiusB = function (edgeIndex) {
        var b = edges[edgeIndex];
        locE.unit = structure.unitMap.get(b.unitB);
        locE.element = locE.unit.elements[b.indexB];
        return theme.size.size(locE) * sizeFactor;
    };
    return {
        linkCount: edgeCount,
        referencePosition: function (edgeIndex) {
            var b = edges[edgeIndex];
            var unitA, unitB;
            var indexA, indexB;
            if (b.unitA < b.unitB) {
                unitA = structure.unitMap.get(b.unitA);
                unitB = structure.unitMap.get(b.unitB);
                indexA = b.indexA;
                indexB = b.indexB;
            }
            else if (b.unitA > b.unitB) {
                unitA = structure.unitMap.get(b.unitB);
                unitB = structure.unitMap.get(b.unitA);
                indexA = b.indexB;
                indexB = b.indexA;
            }
            else {
                throw new Error('same units in createInterUnitBondCylinderMesh');
            }
            return setRefPosition(tmpRef, structure, unitA, indexA) || setRefPosition(tmpRef, structure, unitB, indexB);
        },
        position: function (posA, posB, edgeIndex) {
            var b = edges[edgeIndex];
            var uA = structure.unitMap.get(b.unitA);
            var uB = structure.unitMap.get(b.unitB);
            var rA = radiusA(edgeIndex), rB = radiusB(edgeIndex);
            var r = Math.min(rA, rB) * sizeAspectRatio;
            var oA = Math.sqrt(Math.max(0, rA * rA - r * r)) - 0.05;
            var oB = Math.sqrt(Math.max(0, rB * rB - r * r)) - 0.05;
            uA.conformation.position(uA.elements[b.indexA], posA);
            uB.conformation.position(uB.elements[b.indexB], posB);
            if (oA <= 0.01 && oB <= 0.01)
                return;
            linear_algebra_1.Vec3.normalize(delta, linear_algebra_1.Vec3.sub(delta, posB, posA));
            linear_algebra_1.Vec3.scaleAndAdd(posA, posA, delta, oA);
            linear_algebra_1.Vec3.scaleAndAdd(posB, posB, delta, -oB);
        },
        style: function (edgeIndex) {
            var o = edges[edgeIndex].props.order;
            var f = mol_util_1.BitFlags.create(edges[edgeIndex].props.flag);
            if (types_1.BondType.is(f, 2 /* MetallicCoordination */) || types_1.BondType.is(f, 4 /* HydrogenBond */)) {
                // show metall coordinations and hydrogen bonds with dashed cylinders
                return 1 /* Dashed */;
            }
            else if (o === 2) {
                return 2 /* Double */;
            }
            else if (o === 3) {
                return 3 /* Triple */;
            }
            else {
                return 0 /* Solid */;
            }
        },
        radius: function (edgeIndex) {
            return radius(edgeIndex) * sizeAspectRatio;
        },
        ignore: bond_1.makeInterBondIgnoreTest(structure, props)
    };
}
function createInterUnitBondCylinderImpostors(ctx, structure, theme, props, cylinders) {
    if (!structure.interUnitBonds.edgeCount)
        return cylinders_1.Cylinders.createEmpty(cylinders);
    var builderProps = getInterUnitBondCylinderBuilderProps(structure, theme, props);
    var m = link_1.createLinkCylinderImpostors(ctx, builderProps, props, cylinders);
    var sphere = geometry_1.Sphere3D.expand(geometry_1.Sphere3D(), structure.boundary.sphere, 1 * props.sizeFactor);
    m.setBoundingSphere(sphere);
    return m;
}
function createInterUnitBondCylinderMesh(ctx, structure, theme, props, mesh) {
    if (!structure.interUnitBonds.edgeCount)
        return mesh_1.Mesh.createEmpty(mesh);
    var builderProps = getInterUnitBondCylinderBuilderProps(structure, theme, props);
    var m = link_1.createLinkCylinderMesh(ctx, builderProps, props, mesh);
    var sphere = geometry_1.Sphere3D.expand(geometry_1.Sphere3D(), structure.boundary.sphere, 1 * props.sizeFactor);
    m.setBoundingSphere(sphere);
    return m;
}
exports.InterUnitBondCylinderParams = tslib_1.__assign(tslib_1.__assign(tslib_1.__assign(tslib_1.__assign({}, complex_visual_1.ComplexMeshParams), complex_visual_1.ComplexCylindersParams), bond_1.BondCylinderParams), { sizeFactor: param_definition_1.ParamDefinition.Numeric(0.3, { min: 0, max: 10, step: 0.01 }), sizeAspectRatio: param_definition_1.ParamDefinition.Numeric(2 / 3, { min: 0, max: 3, step: 0.01 }), tryUseImpostor: param_definition_1.ParamDefinition.Boolean(true) });
function InterUnitBondCylinderVisual(materialId, structure, props, webgl) {
    return props.tryUseImpostor && webgl && webgl.extensions.fragDepth
        ? InterUnitBondCylinderImpostorVisual(materialId)
        : InterUnitBondCylinderMeshVisual(materialId);
}
exports.InterUnitBondCylinderVisual = InterUnitBondCylinderVisual;
function InterUnitBondCylinderImpostorVisual(materialId) {
    return complex_visual_1.ComplexCylindersVisual({
        defaultProps: param_definition_1.ParamDefinition.getDefaultValues(exports.InterUnitBondCylinderParams),
        createGeometry: createInterUnitBondCylinderImpostors,
        createLocationIterator: bond_1.BondIterator.fromStructure,
        getLoci: bond_1.getInterBondLoci,
        eachLocation: bond_1.eachInterBond,
        setUpdateState: function (state, newProps, currentProps) {
            state.createGeometry = (newProps.linkScale !== currentProps.linkScale ||
                newProps.linkSpacing !== currentProps.linkSpacing ||
                newProps.ignoreHydrogens !== currentProps.ignoreHydrogens ||
                newProps.linkCap !== currentProps.linkCap ||
                newProps.dashCount !== currentProps.dashCount ||
                newProps.dashScale !== currentProps.dashScale ||
                newProps.dashCap !== currentProps.dashCap ||
                !mol_util_1.arrayEqual(newProps.includeTypes, currentProps.includeTypes) ||
                !mol_util_1.arrayEqual(newProps.excludeTypes, currentProps.excludeTypes));
        },
        mustRecreate: function (structure, props, webgl) {
            return !props.tryUseImpostor || !webgl;
        }
    }, materialId);
}
exports.InterUnitBondCylinderImpostorVisual = InterUnitBondCylinderImpostorVisual;
function InterUnitBondCylinderMeshVisual(materialId) {
    return complex_visual_1.ComplexMeshVisual({
        defaultProps: param_definition_1.ParamDefinition.getDefaultValues(exports.InterUnitBondCylinderParams),
        createGeometry: createInterUnitBondCylinderMesh,
        createLocationIterator: bond_1.BondIterator.fromStructure,
        getLoci: bond_1.getInterBondLoci,
        eachLocation: bond_1.eachInterBond,
        setUpdateState: function (state, newProps, currentProps) {
            state.createGeometry = (newProps.sizeFactor !== currentProps.sizeFactor ||
                newProps.sizeAspectRatio !== currentProps.sizeAspectRatio ||
                newProps.radialSegments !== currentProps.radialSegments ||
                newProps.linkScale !== currentProps.linkScale ||
                newProps.linkSpacing !== currentProps.linkSpacing ||
                newProps.ignoreHydrogens !== currentProps.ignoreHydrogens ||
                newProps.linkCap !== currentProps.linkCap ||
                newProps.dashCount !== currentProps.dashCount ||
                newProps.dashScale !== currentProps.dashScale ||
                newProps.dashCap !== currentProps.dashCap ||
                !mol_util_1.arrayEqual(newProps.includeTypes, currentProps.includeTypes) ||
                !mol_util_1.arrayEqual(newProps.excludeTypes, currentProps.excludeTypes));
        },
        mustRecreate: function (structure, props, webgl) {
            return props.tryUseImpostor && !!webgl;
        }
    }, materialId);
}
exports.InterUnitBondCylinderMeshVisual = InterUnitBondCylinderMeshVisual;
//# sourceMappingURL=bond-inter-unit-cylinder.js.map