/**
 * Copyright (c) 2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * Inspired by https://github.com/dgasmith/gau2grid.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 */
import { __awaiter, __generator, __spreadArrays } from "tslib";
import { sortArray } from '../../mol-data/util';
import { Box3D } from '../../mol-math/geometry';
import { Mat4, Tensor, Vec3 } from '../../mol-math/linear-algebra';
import { Task } from '../../mol-task';
import { arrayMax, arrayMin, arrayRms } from '../../mol-util/array';
import { sphericalCollocation } from './collocation';
import { canComputeAlphaOrbitalsOnGPU, gpuComputeAlphaOrbitalsGridValues } from './gpu/compute';
export function createSphericalCollocationGrid(params, webgl) {
    var _this = this;
    return Task.create('Spherical Collocation Grid', function (ctx) { return __awaiter(_this, void 0, void 0, function () {
        var centers, cParams, matrix;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    centers = params.basis.atoms.map(function (a) { return a.center; });
                    cParams = {
                        grid: initBox(centers, params.gridSpacing, params.boxExpand),
                        basis: params.basis,
                        alphaOrbitals: params.alphaOrbitals,
                        cutoffThreshold: params.cutoffThreshold,
                        sphericalOrder: params.sphericalOrder
                    };
                    if (!canComputeAlphaOrbitalsOnGPU(webgl)) return [3 /*break*/, 1];
                    // console.time('gpu');
                    matrix = gpuComputeAlphaOrbitalsGridValues(webgl, cParams);
                    return [3 /*break*/, 3];
                case 1: return [4 /*yield*/, sphericalCollocation(cParams, ctx)];
                case 2:
                    // console.time('cpu');
                    matrix = _a.sent();
                    _a.label = 3;
                case 3: return [2 /*return*/, createCubeGrid(cParams.grid, matrix, [0, 1, 2], !params.doNotComputeIsovalues)];
            }
        });
    }); });
}
var BohrToAngstromFactor = 0.529177210859;
function createCubeGrid(gridInfo, values, axisOrder, computeIsovalues) {
    var boxSize = Box3D.size(Vec3(), gridInfo.box);
    var boxOrigin = Vec3.clone(gridInfo.box.min);
    Vec3.scale(boxSize, boxSize, BohrToAngstromFactor);
    Vec3.scale(boxOrigin, boxOrigin, BohrToAngstromFactor);
    var scale = Mat4.fromScaling(Mat4(), Vec3.div(Vec3(), boxSize, Vec3.sub(Vec3(), gridInfo.dimensions, Vec3.create(1, 1, 1))));
    var translate = Mat4.fromTranslation(Mat4(), boxOrigin);
    var matrix = Mat4.mul(Mat4(), translate, scale);
    var grid = {
        transform: { kind: 'matrix', matrix: matrix },
        cells: Tensor.create(Tensor.Space(gridInfo.dimensions, axisOrder, Float32Array), values),
        stats: {
            min: arrayMin(values),
            max: arrayMax(values),
            mean: arrayMax(values),
            sigma: arrayRms(values),
        },
    };
    // TODO: when using GPU rendering, the cumulative sum can be computed
    //       along the ray on the fly?
    var isovalues;
    if (computeIsovalues) {
        isovalues = computeIsocontourValues(values, 0.85);
    }
    return { grid: grid, isovalues: isovalues };
}
function initBox(geometry, spacing, expand) {
    var count = geometry.length;
    var box = Box3D.expand(Box3D(), Box3D.fromVec3Array(Box3D(), geometry), Vec3.create(expand, expand, expand));
    var size = Box3D.size(Vec3(), box);
    var spacingThresholds = typeof spacing === 'number' ? [[0, spacing]] : __spreadArrays(spacing);
    spacingThresholds.sort(function (a, b) { return b[0] - a[0]; });
    var s = 0.4;
    for (var i = 0; i <= spacingThresholds.length; i++) {
        s = spacingThresholds[i][1];
        if (spacingThresholds[i][0] <= count)
            break;
    }
    var dimensions = Vec3.ceil(Vec3(), Vec3.scale(Vec3(), size, 1 / s));
    return {
        box: box,
        dimensions: dimensions,
        size: size,
        npoints: dimensions[0] * dimensions[1] * dimensions[2],
        delta: Vec3.div(Vec3(), size, Vec3.subScalar(Vec3(), dimensions, 1)),
    };
}
export function computeIsocontourValues(input, cumulativeThreshold) {
    var weightSum = 0;
    for (var i = 0, _i = input.length; i < _i; i++) {
        var v = input[i];
        var w = v * v;
        weightSum += w;
    }
    var avgWeight = weightSum / input.length;
    var minWeight = 3 * avgWeight;
    // do not try to identify isovalues for degenerate data
    // e.g. all values are almost same
    if (Math.abs(avgWeight - input[0] * input[0]) < 1e-5) {
        return { negative: void 0, positive: void 0 };
    }
    var size = 0;
    while (true) {
        var csum = 0;
        size = 0;
        for (var i = 0, _i = input.length; i < _i; i++) {
            var v = input[i];
            var w = v * v;
            if (w >= minWeight) {
                csum += w;
                size++;
            }
        }
        if (csum / weightSum > cumulativeThreshold) {
            break;
        }
        minWeight -= avgWeight;
    }
    var values = new Float32Array(size);
    var weights = new Float32Array(size);
    var indices = new Int32Array(size);
    var o = 0;
    for (var i = 0, _i = input.length; i < _i; i++) {
        var v = input[i];
        var w = v * v;
        if (w >= minWeight) {
            values[o] = v;
            weights[o] = w;
            indices[o] = o;
            o++;
        }
    }
    sortArray(indices, function (indices, i, j) { return weights[indices[j]] - weights[indices[i]]; });
    var cweight = 0, cutoffIndex = 0;
    for (var i = 0; i < size; i++) {
        cweight += weights[indices[i]];
        if (cweight / weightSum >= cumulativeThreshold) {
            cutoffIndex = i;
            break;
        }
    }
    var positive = Number.POSITIVE_INFINITY, negative = Number.NEGATIVE_INFINITY;
    for (var i = 0; i < cutoffIndex; i++) {
        var v = values[indices[i]];
        if (v > 0) {
            if (v < positive)
                positive = v;
        }
        else if (v < 0) {
            if (v > negative)
                negative = v;
        }
    }
    return {
        negative: negative !== Number.NEGATIVE_INFINITY ? negative : void 0,
        positive: positive !== Number.POSITIVE_INFINITY ? positive : void 0,
    };
}
//# sourceMappingURL=cubes.js.map