/**
 * Copyright (c) 2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
import { __assign, __extends } from "tslib";
import { ParamDefinition as PD } from '../../mol-util/param-definition';
import { PluginBehavior } from '../../mol-plugin/behavior/behavior';
import { Loci, EmptyLoci } from '../../mol-model/loci';
import { ButtonsType, ModifiersKeys } from '../../mol-util/input/input-observer';
import { Binding } from '../../mol-util/binding';
import { PluginStateTransform, PluginStateObject as SO } from '../../mol-plugin-state/objects';
import { Vec3, Mat3 } from '../../mol-math/linear-algebra';
import { deepEqual, UUID } from '../../mol-util';
import { StateTransformer, StateSelection } from '../../mol-state';
import { StructureElement, Model } from '../../mol-model/structure';
import { OrderedSet, SortedArray } from '../../mol-data/int';
import { ObjectControls } from '../../mol-canvas3d/controls/object';
var B = ButtonsType;
var M = ModifiersKeys;
var Trigger = Binding.Trigger;
var DefaultMoveLociBindings = {
    dragMove: Binding([
        Trigger(1 /* Primary */, M.create({ alt: true })),
    ], 'Move Loci', 'Drag using ${triggers}'),
};
var MoveLociParams = {
    bindings: PD.Value(DefaultMoveLociBindings, { isHidden: true }),
};
export var MoveLoci = PluginBehavior.create({
    name: 'move-loci',
    category: 'interaction',
    display: { name: 'Move Loci on Canvas' },
    ctor: /** @class */ (function (_super) {
        __extends(class_1, _super);
        function class_1() {
            return _super !== null && _super.apply(this, arguments) || this;
        }
        class_1.prototype.register = function () {
            var _this = this;
            this.ctx.state.data.actions.add(PatchedModel);
            var busy = false;
            var pendingDir;
            var currentLoci = EmptyLoci;
            var dir = Vec3();
            var m3 = Mat3();
            var ref = Vec3();
            this.subscribeObservable(this.ctx.behaviors.interaction.drag, function (_a) {
                var current = _a.current, buttons = _a.buttons, modifiers = _a.modifiers, pageStart = _a.pageStart, pageEnd = _a.pageEnd;
                if (!_this.ctx.canvas3d)
                    return;
                var binding = _this.params.bindings.dragMove;
                if (!Binding.match(binding, buttons, modifiers))
                    return;
                if (!Loci.areEqual(currentLoci, current.loci)) {
                    currentLoci = current.loci;
                    pendingDir = undefined;
                }
                var loci = Loci.normalize(current.loci, _this.ctx.managers.interactivity.props.granularity);
                if (Loci.isEmpty(loci) || !StructureElement.Loci.is(loci))
                    return;
                Loci.getCenter(loci, ref);
                var direction = ObjectControls.panDirection(Vec3(), pageStart, pageEnd, _this.ctx.canvas3d.camera, ref);
                if (!busy) {
                    var _b = loci.elements[0], unit_1 = _b.unit, indices_1 = _b.indices;
                    var patchedModel = _this.ctx.state.data.select(StateSelection.Generators.ofTransformer(PatchedModel)).filter(function (r) { return r.obj && Model.getRoot(r.obj.data) === Model.getRoot(unit_1.model); })[0];
                    if (!(patchedModel === null || patchedModel === void 0 ? void 0 : patchedModel.obj))
                        return;
                    var u_1 = unit_1.remapModel(patchedModel.obj.data);
                    if (pendingDir) {
                        Vec3.add(dir, pendingDir, direction);
                    }
                    else {
                        Vec3.copy(dir, direction);
                    }
                    // console.log('move', Vec3.magnitude(dir));
                    Vec3.transformMat3(dir, dir, Mat3.directionTransform(m3, u_1.conformation.operator.inverse));
                    var b = _this.ctx.state.data.build();
                    b.to(patchedModel.transform.ref).update(function (old) {
                        old.coords.length = 0;
                        OrderedSet.forEach(indices_1, function (v) {
                            var index = u_1.elements[v];
                            var c = u_1.conformation.invariantPosition(index, Vec3());
                            old.coords.push({ index: index, position: Vec3.add(c, c, dir) });
                        });
                    });
                    busy = true;
                    pendingDir = undefined;
                    b.commit({ canUndo: true, doNotLogTiming: true }).then(function () {
                        busy = false;
                    });
                }
                else if (pendingDir === undefined) {
                    // console.log('pendingPos');
                    pendingDir = direction;
                }
                else {
                    // console.log('wait');
                    Vec3.add(pendingDir, pendingDir, direction);
                }
            });
        };
        class_1.prototype.unregister = function () {
            this.ctx.state.data.actions.remove(PatchedModel);
        };
        return class_1;
    }(PluginBehavior.Handler)),
    params: function () { return MoveLociParams; }
});
var PatchedModel = PluginStateTransform.BuiltIn({
    name: 'patched-model',
    display: { name: 'Patched Model', description: 'Create a patched model.' },
    isDecorator: true,
    from: SO.Molecule.Model,
    to: SO.Molecule.Model,
    params: function (a) {
        return {
            coords: PD.Value([])
        };
    }
})({
    apply: function (_a) {
        var a = _a.a, params = _a.params, cache = _a.cache;
        if (a.data.parent)
            throw new Error('Only the root model can be patched.');
        Model.CoordinatesHistory.set(a.data, new CoordinatesHistory());
        var _b = a.data.atomicConformation, x = _b.x, y = _b.y, z = _b.z;
        var coords = new BufferedCoordinates(x, y, z);
        cache.bufferedCoords = coords;
        if (params.coords.length === 0)
            return new SO.Molecule.Model(a.data);
        return new SO.Molecule.Model(getPatchedModel(a.data, coords.swap(), params.coords));
    },
    update: function (_a) {
        var a = _a.a, b = _a.b, oldParams = _a.oldParams, newParams = _a.newParams, cache = _a.cache;
        if (a.data !== Model.getRoot(b.data))
            return StateTransformer.UpdateResult.Recreate;
        if (deepEqual(oldParams, newParams))
            return StateTransformer.UpdateResult.Unchanged;
        var coords = cache.bufferedCoords;
        b.data = getPatchedModel(b.data, coords.swap(), newParams.coords);
        return StateTransformer.UpdateResult.Updated;
    }
});
var CoordsChangeProp = '__CoordsChange__';
export { CoordsChange };
var CoordsChange = {
    set: function (model, coordsChange) {
        model._staticPropertyData[CoordsChangeProp] = coordsChange;
    },
    get: function (model) {
        return model._staticPropertyData[CoordsChangeProp];
    }
};
var CoordinatesHistory = /** @class */ (function () {
    function CoordinatesHistory() {
    }
    CoordinatesHistory.prototype.areEqual = function (elements, kind, model) {
        var coordsChange = CoordsChange.get(Model.getRoot(model));
        if (coordsChange) {
            return !SortedArray.areIntersecting(elements, coordsChange);
        }
        return false;
    };
    return CoordinatesHistory;
}());
function getPatchedModel(model, coords, patch) {
    var x = coords.x, y = coords.y, z = coords.z;
    var root = Model.getRoot(model);
    var coordsChange = [];
    for (var _i = 0, patch_1 = patch; _i < patch_1.length; _i++) {
        var _a = patch_1[_i], index = _a.index, position = _a.position;
        x[index] = position[0];
        y[index] = position[1];
        z[index] = position[2];
        coordsChange.push(index);
    }
    CoordsChange.set(root, SortedArray.ofUnsortedArray(coordsChange));
    var atomicConformation = __assign(__assign({}, root.atomicConformation), { id: UUID.create22(), x: x, y: y, z: z });
    return __assign(__assign({}, root), { id: UUID.create22(), parent: root, atomicConformation: atomicConformation });
}
var BufferedCoordinates = /** @class */ (function () {
    function BufferedCoordinates(x, y, z) {
        this.x = new BufferedNumberArray(x, Float32Array);
        this.y = new BufferedNumberArray(y, Float32Array);
        this.z = new BufferedNumberArray(z, Float32Array);
    }
    BufferedCoordinates.prototype.swap = function () {
        return {
            x: this.x.swap(),
            y: this.y.swap(),
            z: this.z.swap()
        };
    };
    return BufferedCoordinates;
}());
var BufferedNumberArray = /** @class */ (function () {
    function BufferedNumberArray(orig, ctor) {
        this.orig = orig;
        this.ctor = ctor;
    }
    BufferedNumberArray.prototype.swap = function () {
        if (!this.prev)
            this.prev = new this.ctor(this.orig);
        if (!this.next)
            this.next = new this.ctor(this.prev);
        var ret = this.next;
        ret.set(this.prev);
        this.next = this.prev;
        this.prev = ret;
        return ret;
    };
    return BufferedNumberArray;
}());
//# sourceMappingURL=move-loci.js.map