/**
 * Copyright (c) 2019-2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Áron Samuel Kovács <aron.kovacs@mail.muni.cz>
 */
import { __assign } from "tslib";
/**
 * Adaptive Cutaways for Comprehensible Rendering of Polygonal Scenes by Michael Burns and Adam Finkelstein
 * https://gfx.cs.princeton.edu/pubs/Burns_2008_ACF/adaptive_cutaways.pdf
 */
import { QuadSchema, QuadValues } from '../../mol-gl/compute/util';
import { TextureSpec, UniformSpec } from '../../mol-gl/renderable/schema';
import { ShaderCode } from '../../mol-gl/shader-code';
import { createComputeRenderItem } from '../../mol-gl/webgl/render-item';
import { createComputeRenderable } from '../../mol-gl/renderable';
import { Vec2, Vec3, Vec4 } from '../../mol-math/linear-algebra';
import { ValueCell } from '../../mol-util';
import quad_vert from '../../mol-gl/shader/quad.vert';
import cutaway_jfa_step_frag from '../../mol-gl/shader/cutaway-jfa-step.frag';
import cutaway_init_frag from '../../mol-gl/shader/cutaway-init.frag';
import cutaway_copy_to_target_frag from '../../mol-gl/shader/cutaway-copy-to-target.frag';
import { ParamDefinition as PD } from '../../mol-util/param-definition';
import { StereoCamera } from '../camera/stereo';
import { isDebugMode } from '../../mol-util/debug';
export var CutawayParams = {
    angle: PD.Numeric(45, { min: 0, max: 90, step: 1 }),
    borderSize: PD.Numeric(0.05, { min: 0, max: 0.1, step: 0.01 }),
};
var CutawayInitSchema = __assign(__assign({}, QuadSchema), { tDepth: TextureSpec('texture', 'rgba', 'ubyte', 'nearest'), uTexSize: UniformSpec('v2'), uIsOrtho: UniformSpec('f'), uNear: UniformSpec('f'), uFar: UniformSpec('f') });
function getCutawayInitRenderable(ctx, depthTexture) {
    var values = __assign(__assign({}, QuadValues), { tDepth: ValueCell.create(depthTexture), uTexSize: ValueCell.create(Vec2.create(depthTexture.getWidth(), depthTexture.getHeight())), uIsOrtho: ValueCell.create(0.0), uNear: ValueCell.create(0.0), uFar: ValueCell.create(1.0) });
    var schema = __assign({}, CutawayInitSchema);
    var shaderCode = ShaderCode('cutaway-init', quad_vert, cutaway_init_frag);
    var renderItem = createComputeRenderItem(ctx, 'triangles', shaderCode, schema, values);
    return createComputeRenderable(renderItem, values);
}
var CutawayJfaSchema = __assign(__assign({}, QuadSchema), { tCutaway: TextureSpec('texture', 'rgba', 'float', 'nearest'), uTexSize: UniformSpec('v2'), uViewport: UniformSpec('v4'), uStep: UniformSpec('v2'), uPMsz: UniformSpec('f'), uAngle: UniformSpec('f'), uEdgeRegionSize: UniformSpec('f'), uAspectRatio: UniformSpec('v2'), uNear: UniformSpec('f'), uFar: UniformSpec('f') });
function getCutawayJfaRenderable(ctx, cutawayTexture) {
    var values = __assign(__assign({}, QuadValues), { tCutaway: ValueCell.create(cutawayTexture), uTexSize: ValueCell.create(Vec2.create(cutawayTexture.getWidth(), cutawayTexture.getHeight())), uViewport: ValueCell.create(Vec4.create(0.0, 0.0, 1.0, 1.0)), uStep: ValueCell.create(Vec2.create(0, 0)), uPMsz: ValueCell.create(1.0), uAngle: ValueCell.create(Math.PI / 3.0), uEdgeRegionSize: ValueCell.create(0.0), uAspectRatio: ValueCell.create(Vec2.create(1.0, 1.0)), uNear: ValueCell.create(0.0), uFar: ValueCell.create(1.0) });
    var schema = __assign({}, CutawayJfaSchema);
    var shaderCode = ShaderCode('cutaway-jfa', quad_vert, cutaway_jfa_step_frag);
    var renderItem = createComputeRenderItem(ctx, 'triangles', shaderCode, schema, values);
    return createComputeRenderable(renderItem, values);
}
var CutawayCopyToTargetSchema = __assign(__assign({}, QuadSchema), { tDepth: TextureSpec('texture', 'rgba', 'float', 'nearest'), uTexSize: UniformSpec('v2'), uIsOrtho: UniformSpec('f'), uNear: UniformSpec('f'), uFar: UniformSpec('f') });
function getCutawayCopyToTargetRenderable(ctx, depthTexture) {
    var values = __assign(__assign({}, QuadValues), { tDepth: ValueCell.create(depthTexture), uTexSize: ValueCell.create(Vec2.create(depthTexture.getWidth(), depthTexture.getHeight())), uIsOrtho: ValueCell.create(0.0), uNear: ValueCell.create(0.0), uFar: ValueCell.create(1.0) });
    var schema = __assign({}, CutawayCopyToTargetSchema);
    var shaderCode = ShaderCode('cutaway-copy-to-target', quad_vert, cutaway_copy_to_target_frag);
    var renderItem = createComputeRenderItem(ctx, 'triangles', shaderCode, schema, values);
    return createComputeRenderable(renderItem, values);
}
var CutawayPass = /** @class */ (function () {
    function CutawayPass(webgl, width, height) {
        this.webgl = webgl;
        this._supported = CutawayPass.isSupported(webgl);
        if (!this.supported) {
            return;
        }
        this._target = webgl.createRenderTarget(width, height, true, 'uint8', 'nearest');
        this.jfaATarget = webgl.createRenderTarget(width, height, false, 'float32', 'nearest');
        this.jfaBTarget = webgl.createRenderTarget(width, height, false, 'float32', 'nearest');
        this.jfaARenderable = getCutawayJfaRenderable(webgl, this.jfaATarget.texture);
        this.jfaBRenderable = getCutawayJfaRenderable(webgl, this.jfaBTarget.texture);
        this.initRenderable = getCutawayInitRenderable(webgl, this._target.texture);
        this.copyToTargetRenderable = getCutawayCopyToTargetRenderable(webgl, this.jfaATarget.texture);
    }
    Object.defineProperty(CutawayPass.prototype, "supported", {
        get: function () {
            return this._supported;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(CutawayPass.prototype, "target", {
        get: function () {
            return this.supported ? this._target.texture : null;
        },
        enumerable: false,
        configurable: true
    });
    CutawayPass.isSupported = function (webgl) {
        var _a = webgl.extensions, textureFloat = _a.textureFloat, colorBufferFloat = _a.colorBufferFloat;
        if (!textureFloat || !colorBufferFloat) {
            if (isDebugMode) {
                var missing = [];
                if (!textureFloat)
                    missing.push('textureFloat');
                if (!colorBufferFloat)
                    missing.push('colorBufferFloat');
                console.log("Missing \"" + missing.join('", "') + "\" extensions required for \"cutaway\"");
            }
            return false;
        }
        else {
            return true;
        }
    };
    CutawayPass.prototype.setRenderDepthState = function (renderer, camera) {
        var _a = this.webgl, gl = _a.gl, state = _a.state;
        var _b = camera.viewport, x = _b.x, y = _b.y, width = _b.width, height = _b.height;
        renderer.setDrawingBufferSize(this._target.getWidth(), this._target.getHeight());
        renderer.setViewport(x, y, width, height);
        renderer.update(camera);
        gl.viewport(x, y, width, height);
        gl.scissor(x, y, width, height);
        state.enable(gl.SCISSOR_TEST);
        state.enable(gl.DEPTH_TEST);
        state.colorMask(true, true, true, true);
        state.clearColor(1, 1, 1, 1);
        state.depthMask(true);
        gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
    };
    CutawayPass.prototype.setViewport = function (viewport) {
        ValueCell.update(this.jfaARenderable.values.uViewport, Vec4.copy(this.jfaARenderable.values.uViewport.ref.value, viewport));
        ValueCell.update(this.jfaBRenderable.values.uViewport, Vec4.copy(this.jfaBRenderable.values.uViewport.ref.value, viewport));
    };
    CutawayPass.prototype.setJfaState = function (camera, scene, props) {
        var _a = this.webgl, gl = _a.gl, state = _a.state;
        var orthographic = camera.state.mode === 'orthographic' ? 1 : 0;
        state.enable(gl.SCISSOR_TEST);
        state.disable(gl.BLEND);
        state.disable(gl.DEPTH_TEST);
        state.depthMask(false);
        var PMsz = orthographic ? 0 : (camera.near + camera.far) / (camera.near - camera.far);
        ValueCell.updateIfChanged(this.jfaARenderable.values.uPMsz, PMsz);
        ValueCell.updateIfChanged(this.jfaBRenderable.values.uPMsz, PMsz);
        ValueCell.updateIfChanged(this.initRenderable.values.uIsOrtho, orthographic);
        ValueCell.updateIfChanged(this.copyToTargetRenderable.values.uIsOrtho, orthographic);
        ValueCell.updateIfChanged(this.initRenderable.values.uNear, camera.near);
        ValueCell.updateIfChanged(this.jfaARenderable.values.uNear, camera.near);
        ValueCell.updateIfChanged(this.jfaBRenderable.values.uNear, camera.near);
        ValueCell.updateIfChanged(this.copyToTargetRenderable.values.uNear, camera.near);
        ValueCell.updateIfChanged(this.initRenderable.values.uFar, camera.far);
        ValueCell.updateIfChanged(this.jfaARenderable.values.uFar, camera.far);
        ValueCell.updateIfChanged(this.jfaBRenderable.values.uFar, camera.far);
        ValueCell.updateIfChanged(this.copyToTargetRenderable.values.uFar, camera.far);
        var angle = props.angle * Math.PI / 180.0;
        ValueCell.updateIfChanged(this.jfaARenderable.values.uAngle, angle);
        ValueCell.updateIfChanged(this.jfaBRenderable.values.uAngle, angle);
        ValueCell.updateIfChanged(this.jfaARenderable.values.uEdgeRegionSize, props.borderSize);
        ValueCell.updateIfChanged(this.jfaBRenderable.values.uEdgeRegionSize, props.borderSize);
        var aspectRatio = this.getScalingRatio(camera, scene);
        ValueCell.update(this.jfaARenderable.values.uAspectRatio, Vec2.copy(this.jfaARenderable.values.uAspectRatio.ref.value, aspectRatio));
        ValueCell.update(this.jfaBRenderable.values.uAspectRatio, Vec2.copy(this.jfaBRenderable.values.uAspectRatio.ref.value, aspectRatio));
    };
    CutawayPass.prototype.getScalingRatio = function (camera, scene) {
        var distanceToCamera = Vec3.distance(camera.state.position, scene.boundingSphereVisible.center);
        var center = Vec4.create(0, 0, distanceToCamera, 1);
        var right = Vec4.create(scene.boundingSphereVisible.radius, 0, distanceToCamera, 1);
        var top = Vec4.create(0, scene.boundingSphereVisible.radius, distanceToCamera, 1);
        // project
        Vec4.transformMat4(center, center, camera.projection);
        Vec4.transformMat4(right, right, camera.projection);
        Vec4.transformMat4(top, top, camera.projection);
        // scale only the relevant coordinates
        center[0] = center[0] / center[3];
        center[1] = center[1] / center[3];
        right[0] = right[0] / right[3];
        top[1] = top[1] / top[3];
        // distances between points
        var ratio = Vec2.create(Math.abs(right[0] - center[0]), Math.abs(top[1] - center[1]));
        // NDC -> screen space
        Vec2.scale(ratio, ratio, 0.5);
        // and we want to scale by the inverse
        Vec2.inverse(ratio, ratio);
        Vec2.set(ratio, 1, 1);
        return ratio;
    };
    CutawayPass.prototype.renderDepth = function (renderer, camera, scene) {
        this._target.bind();
        this.setRenderDepthState(renderer, camera);
        var ligandId = 27;
        var primitives = scene.primitives.renderables.concat();
        var ligand = primitives.filter(function (r) { return r.id === ligandId; });
        var groupLigand = {
            renderables: ligand,
            direction: scene.primitives.direction,
            position: scene.primitives.position,
            up: scene.primitives.up,
            view: scene.primitives.view
        };
        renderer.renderDepth(groupLigand, camera, null, null);
    };
    CutawayPass.prototype.renderJfa = function (camera, scene, props) {
        this.setJfaState(camera, scene, props);
        // initialize the A JFA buffer
        this.jfaATarget.bind();
        this.initRenderable.render();
        // JFA
        var width = this.jfaATarget.getWidth();
        var height = this.jfaATarget.getHeight();
        var stepNormalized = Vec2();
        var readingA = true;
        var stepPixels = Math.ceil(Math.max(width, height) * 0.5);
        while (true) {
            Vec2.set(stepNormalized, stepPixels / width, stepPixels / height);
            if (readingA) {
                this.jfaBTarget.bind();
                ValueCell.update(this.jfaARenderable.values.uStep, stepNormalized);
                this.jfaARenderable.render();
            }
            else {
                this.jfaATarget.bind();
                ValueCell.update(this.jfaBRenderable.values.uStep, stepNormalized);
                this.jfaBRenderable.render();
            }
            readingA = !readingA;
            if (stepPixels <= 1) {
                break;
            }
            stepPixels = Math.ceil(stepPixels / 2);
        }
        // copy from the final JFA buffer
        var finalJfaTexture = readingA ? this.jfaATarget.texture : this.jfaBTarget.texture;
        if (this.copyToTargetRenderable.values.tDepth.ref.value !== finalJfaTexture) {
            ValueCell.update(this.copyToTargetRenderable.values.tDepth, finalJfaTexture);
            this.copyToTargetRenderable.update();
        }
        this._target.bind();
        this.copyToTargetRenderable.render();
    };
    CutawayPass.prototype.render = function (renderer, camera, scene, props) {
        if (!this.supported) {
            return;
        }
        if (StereoCamera.is(camera)) {
            this.setViewport(Vec4.create(0, 0, 0.5, 1));
            this.renderDepth(renderer, camera.left, scene);
            this.renderJfa(camera.left, scene, props);
            this.setViewport(Vec4.create(0.5, 0, 0.5, 1));
            this.renderDepth(renderer, camera.right, scene);
            this.renderJfa(camera.right, scene, props);
        }
        else {
            this.setViewport(Vec4.create(0, 0, 1, 1));
            this.renderDepth(renderer, camera, scene);
            this.renderJfa(camera, scene, props);
        }
    };
    CutawayPass.prototype.setSize = function (width, height) {
        var _a = this.jfaARenderable.values.uTexSize.ref.value, w = _a[0], h = _a[1];
        if (width !== w || height !== h) {
            this._target.setSize(width, height);
            this.jfaATarget.setSize(width, height);
            this.jfaBTarget.setSize(width, height);
            ValueCell.update(this.initRenderable.values.uTexSize, Vec2.set(this.initRenderable.values.uTexSize.ref.value, width, height));
            ValueCell.update(this.jfaARenderable.values.uTexSize, Vec2.set(this.jfaARenderable.values.uTexSize.ref.value, width, height));
            ValueCell.update(this.jfaBRenderable.values.uTexSize, Vec2.set(this.jfaBRenderable.values.uTexSize.ref.value, width, height));
            ValueCell.update(this.copyToTargetRenderable.values.uTexSize, Vec2.set(this.copyToTargetRenderable.values.uTexSize.ref.value, width, height));
        }
    };
    return CutawayPass;
}());
export { CutawayPass };
//# sourceMappingURL=cutaway.js.map