/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 * @author David Sehnal <david.sehnal@gmail.com>
 */
import { defaults, noop } from '../../mol-util';
import { SimpleBuffer } from './simple-buffer';
// only import 'fs' in node.js
var fs = typeof document === 'undefined' ? require('fs') : void 0;
export var FileHandle;
(function (FileHandle) {
    function fromBuffer(buffer, name) {
        return {
            name: name,
            readBuffer: function (position, sizeOrBuffer, size, byteOffset) {
                var bytesRead;
                var outBuffer;
                if (typeof sizeOrBuffer === 'number') {
                    size = defaults(size, sizeOrBuffer);
                    var start = position;
                    var end = Math.min(buffer.length, start + size);
                    bytesRead = end - start;
                    outBuffer = SimpleBuffer.fromUint8Array(new Uint8Array(buffer.buffer, start, end - start));
                }
                else {
                    size = defaults(size, sizeOrBuffer.length);
                    var start = position;
                    var end = Math.min(buffer.length, start + size);
                    sizeOrBuffer.set(buffer.subarray(start, end), byteOffset);
                    bytesRead = end - start;
                    outBuffer = sizeOrBuffer;
                }
                if (size !== bytesRead) {
                    console.warn("byteCount " + size + " and bytesRead " + bytesRead + " differ");
                }
                return Promise.resolve({ bytesRead: bytesRead, buffer: outBuffer });
            },
            writeBuffer: function (position, buffer, length) {
                length = defaults(length, buffer.length);
                console.error('.writeBuffer not implemented for FileHandle.fromBuffer');
                return Promise.resolve(0);
            },
            writeBufferSync: function (position, buffer, length) {
                length = defaults(length, buffer.length);
                console.error('.writeSync not implemented for FileHandle.fromBuffer');
                return 0;
            },
            close: noop
        };
    }
    FileHandle.fromBuffer = fromBuffer;
    function fromDescriptor(file, name) {
        if (fs === undefined)
            throw new Error('fs module not available');
        return {
            name: name,
            readBuffer: function (position, sizeOrBuffer, length, byteOffset) {
                return new Promise(function (res, rej) {
                    var outBuffer;
                    if (typeof sizeOrBuffer === 'number') {
                        byteOffset = defaults(byteOffset, 0);
                        length = defaults(length, sizeOrBuffer);
                        outBuffer = SimpleBuffer.fromArrayBuffer(new ArrayBuffer(sizeOrBuffer));
                    }
                    else {
                        byteOffset = defaults(byteOffset, 0);
                        length = defaults(length, sizeOrBuffer.length);
                        outBuffer = sizeOrBuffer;
                    }
                    fs.read(file, outBuffer, byteOffset, length, position, function (err, bytesRead, buffer) {
                        if (err) {
                            rej(err);
                            return;
                        }
                        if (length !== bytesRead) {
                            console.warn("byteCount " + length + " and bytesRead " + bytesRead + " differ");
                        }
                        res({ bytesRead: bytesRead, buffer: buffer });
                    });
                });
            },
            writeBuffer: function (position, buffer, length) {
                length = defaults(length, buffer.length);
                return new Promise(function (res, rej) {
                    fs.write(file, buffer, 0, length, position, function (err, written) {
                        if (err)
                            rej(err);
                        else
                            res(written);
                    });
                });
            },
            writeBufferSync: function (position, buffer, length) {
                length = defaults(length, buffer.length);
                return fs.writeSync(file, buffer, 0, length, position);
            },
            close: function () {
                try {
                    if (file !== void 0)
                        fs.close(file, noop);
                }
                catch (e) {
                }
            }
        };
    }
    FileHandle.fromDescriptor = fromDescriptor;
})(FileHandle || (FileHandle = {}));
//# sourceMappingURL=file-handle.js.map